(*
  Implementao do Algoritmo de Inteligncia Artificial A* (A Estrela)
  Autor: Diego Silva
  Contato: drsilva@gmail.com 
*)
unit UMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Grids, StdCtrls, Buttons, ColorGrd, ExtCtrls, DSUtil, Spin;

type
  TCelula = Record
    Posicao: TPoint;
    Custo: Double;
    CustoTerreno: Double;
    Diagonal: Boolean;
    G: Double;
    H: Double;
    Pai: Integer;
  end;

  TLista = array of TCelula;
  TfrmMain = class(TForm)
    Mapa: TStringGrid;
    SaveDialog1: TSaveDialog;
    OpenDialog1: TOpenDialog;
    Panel1: TPanel;
    BitBtn1: TBitBtn;
    Button1: TButton;
    pnlAgua: TPanel;
    pnlBarro: TPanel;
    pnlBarreira: TPanel;
    BitBtn2: TBitBtn;
    BitBtn3: TBitBtn;
    BitBtn4: TBitBtn;
    BitBtn5: TBitBtn;
    pnlDestino: TPanel;
    pnlInicio: TPanel;
    BitBtn6: TBitBtn;
    pnlLimpar: TPanel;
    BitBtn7: TBitBtn;
    BitBtn8: TBitBtn;
    BitBtn9: TBitBtn;
    speLinhas: TSpinEdit;
    speColunas: TSpinEdit;
    speTamanho: TSpinEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    lblStatus: TLabel;
    speVelocidade: TSpinEdit;
    Label4: TLabel;
    procedure MapaDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure MapaClick(Sender: TObject);
    procedure MapaDblClick(Sender: TObject);
    procedure BitBtn5Click(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure BitBtn7Click(Sender: TObject);
    procedure BitBtn6Click(Sender: TObject);
    procedure BitBtn8Click(Sender: TObject);
    procedure BitBtn10Click(Sender: TObject);
    procedure BitBtn9Click(Sender: TObject);
    procedure BitBtn11Click(Sender: TObject);
    procedure BitBtn12Click(Sender: TObject);
    procedure BitBtn13Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure speLinhasChange(Sender: TObject);
    procedure speColunasChange(Sender: TObject);
    procedure speTamanhoChange(Sender: TObject);
  private
    { Private declarations }
    CorSelecionada: TColor;

    AchouCaminho: Boolean;
    ListaAberta: TLista;
    ListaFechada: TLista;
    Caminho: TLista;
    MelhorCaminho: TLista;

    CelulaAtual: TCelula;

    function DireitaLiberado: Boolean;
    function EsquerdaLiberado: Boolean;
    function AbaixoLiberado: Boolean;
    function AcimaLiberado: Boolean;

    function DireitaSuperiorLiberado: Boolean;
    function EsquerdaSuperiorLiberado: Boolean;
    function DireitaInferiorLiberado: Boolean;
    function EsquerdaInferiorLiberado: Boolean;

    procedure AStar;
    procedure Pintar(Cor: TColor);
    function GetGoal: TCelula;
    function GetStart: TCelula;
    procedure Adicionar(var Lista: TLista; No: TCelula);
    procedure Remover(var Lista: TLista; No: TCelula);
    function CelulaMenorCusto(var Lista: TLista): TCelula;
    procedure VerificarVizinhos(var cDireita: TCelula; var cEsquerda: TCelula; var cAcima: TCelula; var cAbaixo: TCelula);
    function EstaNaLista(var aLista: TLista; No: TCelula): Boolean;
    function CelulaIgual(Celula1, Celula2: TCelula): Boolean;
    procedure AtualizaCusto(var Atual: TCelula; const Inicial: TCelula; const Final: TCelula);
    procedure ProcessarCelula(var Processar: TCelula; var Inicial: TCelula; var Final: TCelula;  var Atual: TCelula; IndexPai: Integer);
    procedure VerificarDiagonais(var cSD, cSE, cID, cIE: TCelula);
    procedure Animar;
    procedure LimparTudo;
    procedure LimparDiagonais;
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;

const
  CustoBarro = 600;
  CustoAgua = 1200;
  Cor_Visitado = '$00B8FA98';

implementation

uses Math;

{$R *.dfm}

{ TForm1 }

function TfrmMain.AbaixoLiberado: Boolean;
begin
  Result := (CelulaAtual.Posicao.X < Mapa.RowCount-1) and
            (
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y, CelulaAtual.Posicao.X+1]) = '') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y, CelulaAtual.Posicao.X+1]) = 'CLLIME') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y, CelulaAtual.Posicao.X+1]) = 'CLRED') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y, CelulaAtual.Posicao.X+1]) = ColorToString(pnlBarro.Color)) or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y, CelulaAtual.Posicao.X+1]) = ColorToString(pnlAgua.Color))
           );
end;

function TfrmMain.AcimaLiberado: Boolean;
begin
  Result := (CelulaAtual.Posicao.X > 0) and
            (
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y, CelulaAtual.Posicao.X-1]) = '') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y, CelulaAtual.Posicao.X-1]) = 'CLLIME') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y, CelulaAtual.Posicao.X-1]) = 'CLRED') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y, CelulaAtual.Posicao.X-1]) = ColorToString(pnlBarro.Color)) or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y, CelulaAtual.Posicao.X-1]) = ColorToString(pnlAgua.Color))
           );
end;

function TfrmMain.DireitaLiberado: Boolean;
var
  res: Boolean;
begin
  Result := (CelulaAtual.Posicao.Y < Mapa.ColCount -1) and
            (
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y+1, CelulaAtual.Posicao.X]) = '') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y+1, CelulaAtual.Posicao.X]) = 'CLLIME') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y+1, CelulaAtual.Posicao.X]) = 'CLRED') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y+1, CelulaAtual.Posicao.X]) = ColorToString(pnlBarro.Color)) or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y+1, CelulaAtual.Posicao.X]) = ColorToString(pnlAgua.Color))
           );
end;

function TfrmMain.EsquerdaLiberado: Boolean;
begin
  Result := (CelulaAtual.Posicao.Y > 0) and
            (
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y-1, CelulaAtual.Posicao.X]) = '') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y-1, CelulaAtual.Posicao.X]) = 'CLLIME') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y-1, CelulaAtual.Posicao.X]) = 'CLRED') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y-1, CelulaAtual.Posicao.X]) = ColorToString(pnlBarro.Color)) or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y-1, CelulaAtual.Posicao.X]) = ColorToString(pnlAgua.Color))
           );
end;

procedure TfrmMain.MapaDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
begin
  if UpperCase(Mapa.Cells[ACol, ARow]) = 'CLWHITE' then
  begin
    Mapa.Cells[ACol, ARow] := '';

    Mapa.Canvas.Brush.Color := clWhite;
    Mapa.Canvas.FillRect(Rect);
  end;

  if Mapa.Cells[ACol, ARow] <> '' then
  begin
    Mapa.Canvas.Brush.Color := StringToColor(Mapa.Cells[ACol, ARow]);
    Mapa.Canvas.FillRect(Rect);
  end;
end;

procedure TfrmMain.MapaClick(Sender: TObject);
var
  x, y: Integer;
begin
  if (CorSelecionada = pnlInicio.Color) or (CorSelecionada = pnlDestino.Color) then
  begin
    for x := 0 to Mapa.RowCount -1 do
    begin
      for y := 0 to Mapa.ColCount -1 do
      begin
        if Mapa.Cells[y, x] = ColorToString(CorSelecionada) then
          Mapa.Cells[y, x] := '';
      end;
    end;
  end;

  Pintar(CorSelecionada);
  CelulaAtual.Posicao.X := Mapa.Row;
  CelulaAtual.Posicao.Y := Mapa.Col;
end;

procedure TfrmMain.MapaDblClick(Sender: TObject);
begin
  CelulaAtual.Posicao.X := Mapa.Row;
  CelulaAtual.Posicao.Y := Mapa.Col;
end;

procedure TfrmMain.Pintar(Cor: TColor);
var
  X, Y: Integer;
begin
  for Y := Mapa.Selection.Left to Mapa.Selection.Right do
    for X := Mapa.Selection.Top to Mapa.Selection.Bottom do
      Mapa.Cells[Y, X] := ColorToString(Cor);

  Mapa.Repaint;
end;

function TfrmMain.GetStart: TCelula;
var
  x, y: Integer;
  Celula: TCelula;
  Achou: Boolean;
begin
  Achou := False;
  for x := 0 to Mapa.RowCount -1 do
  begin
    for y := 0 to Mapa.ColCount -1 do
    begin
      if AnsiUpperCase(Mapa.Cells[y, x]) = 'CLLIME' then
      begin
        Celula.Posicao.X := x;
        Celula.Posicao.Y := y;
        Achou := True;
        Break;
      end;
    end;

    if Achou then
      Break;
  end;
  Result := Celula;
end;

function TfrmMain.GetGoal: TCelula;
var
  x, y: Integer;
  Celula: TCelula;
  Achou: Boolean;
begin
  Achou := False;
  for x := 0 to Mapa.RowCount -1 do
  begin
    for y := 0 to Mapa.ColCount -1 do
    begin
      if AnsiUpperCase(Mapa.Cells[y, x]) = 'CLRED' then
      begin
        Celula.Posicao.X := x;
        Celula.Posicao.Y := y;
        Achou := True;
        Break;
      end;
    end;

    if Achou then
      Break;
  end;

  Result := Celula;
end;

procedure TfrmMain.Adicionar(var Lista: TLista; No: TCelula);
var
  Tamanho: Integer;
begin
  if not EstaNaLista(Lista, No) then
  begin
    Tamanho := Length(Lista);
    SetLength(Lista, Tamanho+1);
    Lista[Tamanho] := No;
  end;
end;

procedure TfrmMain.VerificarVizinhos(var cDireita: TCelula; var cEsquerda: TCelula; var cAcima: TCelula; var cAbaixo: TCelula);
begin
  cDireita.Pai := -1;
  cEsquerda.Pai := -1;
  cAbaixo.Pai := -1;
  cAcima.Pai := -1;
  cDireita.CustoTerreno := 0;
  cEsquerda.CustoTerreno := 0;
  cAbaixo.CustoTerreno := 0;
  cAcima.CustoTerreno := 0;

  if DireitaLiberado then
  begin
    cDireita.Posicao.X := CelulaAtual.Posicao.X;
    cDireita.Posicao.Y := CelulaAtual.Posicao.Y +1;
    cDireita.G := CelulaAtual.G + 1;

    if Mapa.Cells[cDireita.Posicao.Y, cDireita.Posicao.X] = ColorToString(pnlBarro.Color) then
     cDireita.CustoTerreno := CustoBarro;

    if Mapa.Cells[cDireita.Posicao.Y, cDireita.Posicao.X] = ColorToString(pnlAgua.Color) then
     cDireita.CustoTerreno := CustoAgua;
  end
  else
  begin
    cDireita.Posicao.X := -1;
    cDireita.Posicao.Y := -1;
    cDireita.Custo := -1;
  end;

  if EsquerdaLiberado then
  begin
    cEsquerda.Posicao.X := CelulaAtual.Posicao.X;
    cEsquerda.Posicao.Y := CelulaAtual.Posicao.Y -1;
    cEsquerda.G := CelulaAtual.G + 1;

    if Mapa.Cells[cEsquerda.Posicao.Y, cEsquerda.Posicao.X] = ColorToString(pnlBarro.Color) then
     cEsquerda.CustoTerreno := CustoBarro;

    if Mapa.Cells[cEsquerda.Posicao.Y, cEsquerda.Posicao.X] = ColorToString(pnlAgua.Color) then
     cEsquerda.CustoTerreno := CustoAgua;
  end
  else
  begin
    cEsquerda.Posicao.X := -1;
    cEsquerda.Posicao.Y := -1;
    cEsquerda.Custo := -1;
  end;

  if AcimaLiberado then
  begin
    cAcima.Posicao.X := CelulaAtual.Posicao.X-1;
    cAcima.Posicao.Y := CelulaAtual.Posicao.Y;
    cAcima.G := CelulaAtual.G + 1;

    if Mapa.Cells[cAcima.Posicao.Y, cAcima.Posicao.X] = ColorToString(pnlBarro.Color) then
     cAcima.CustoTerreno := CustoBarro;

    if Mapa.Cells[cAcima.Posicao.Y, cAcima.Posicao.Y] = ColorToString(pnlAgua.Color) then
     cAcima.CustoTerreno := CustoAgua;
  end
  else
  begin
    cAcima.Posicao.X := -1;
    cAcima.Posicao.Y := -1;
    cAcima.Custo := -1;
  end;

  if AbaixoLiberado then
  begin
    cAbaixo.Posicao.X := CelulaAtual.Posicao.X+1;
    cAbaixo.Posicao.Y := CelulaAtual.Posicao.Y;
    cAbaixo.G := CelulaAtual.G + 1;

    if Mapa.Cells[cAbaixo.Posicao.Y, cAbaixo.Posicao.X] = ColorToString(pnlBarro.Color) then
     cAbaixo.CustoTerreno := CustoBarro;

    if Mapa.Cells[cAbaixo.Posicao.Y, cAbaixo.Posicao.X] = ColorToString(pnlAgua.Color) then
     cAbaixo.CustoTerreno := CustoAgua;
  end
  else
  begin
    cAbaixo.Posicao.X := -1;
    cAbaixo.Posicao.Y := -1;
    cAbaixo.Custo := -1;
  end;
end;

procedure TfrmMain.VerificarDiagonais(var cSD: TCelula; var cSE: TCelula; var cID: TCelula; var cIE: TCelula);
begin
  cSD.Pai := -1;
  cSE.Pai := -1;
  cID.Pai := -1;
  cIE.Pai := -1;

  cSD.CustoTerreno := 0;
  cSE.CustoTerreno := 0;
  cID.CustoTerreno := 0;
  cIE.CustoTerreno := 0;

  if DireitaSuperiorLiberado then
  begin
    cSD.Posicao.X := CelulaAtual.Posicao.X -1;
    cSD.Posicao.Y := CelulaAtual.Posicao.Y +1;
    cSD.G := CelulaAtual.G + 1;

    if Mapa.Cells[cSD.Posicao.Y, cSD.Posicao.X] = ColorToString(pnlBarro.Color) then
     cSD.CustoTerreno := CustoBarro;

    if Mapa.Cells[cSD.Posicao.Y, cSD.Posicao.X] = ColorToString(pnlAgua.Color) then
     cSD.CustoTerreno := CustoAgua;
  end
  else
  begin
    cSD.Posicao.X := -1;
    cSD.Posicao.Y := -1;
    cSD.Custo := -1;
  end;

  if EsquerdaSuperiorLiberado then
  begin
    cSE.Posicao.X := CelulaAtual.Posicao.X -1;
    cSE.Posicao.Y := CelulaAtual.Posicao.Y -1;
    cSE.G := CelulaAtual.G + 1;

    if Mapa.Cells[cSE.Posicao.Y, cSE.Posicao.X] = ColorToString(pnlBarro.Color) then
     cSE.CustoTerreno := CustoBarro;

    if Mapa.Cells[cSE.Posicao.Y, cSE.Posicao.X] = ColorToString(pnlAgua.Color) then
     cSE.CustoTerreno := CustoAgua;
  end
  else
  begin
    cSE.Posicao.X := -1;
    cSE.Posicao.Y := -1;
    cSE.Custo := -1;
  end;

  if DireitaInferiorLiberado then
  begin
    cID.Posicao.X := CelulaAtual.Posicao.X+1;
    cID.Posicao.Y := CelulaAtual.Posicao.Y+1;
    cID.G := CelulaAtual.G + 1;

    if Mapa.Cells[cID.Posicao.Y, cID.Posicao.X] = ColorToString(pnlBarro.Color) then
     cID.CustoTerreno := CustoBarro;

    if Mapa.Cells[cID.Posicao.Y, cID.Posicao.X] = ColorToString(pnlAgua.Color) then
     cID.CustoTerreno := CustoAgua;
  end
  else
  begin
    cID.Posicao.X := -1;
    cID.Posicao.Y := -1;
    cID.Custo := -1;
  end;

  if EsquerdaInferiorLiberado then
  begin
    cIE.Posicao.X := CelulaAtual.Posicao.X+1;
    cIE.Posicao.Y := CelulaAtual.Posicao.Y-1;
    cIE.G := CelulaAtual.G + 1;

    if Mapa.Cells[cIE.Posicao.Y, cIE.Posicao.X] = ColorToString(pnlBarro.Color) then
     cIE.CustoTerreno := CustoBarro;

    if Mapa.Cells[cIE.Posicao.Y, cIE.Posicao.X] = ColorToString(pnlAgua.Color) then
     cIE.CustoTerreno := CustoAgua;
  end
  else
  begin
    cIE.Posicao.X := -1;
    cIE.Posicao.Y := -1;
    cIE.Custo := -1;
  end;
end;

procedure TfrmMain.Remover(var Lista: TLista; No: TCelula);
var
  i, Index: Integer;
begin
  for Index := 0 to Length(Lista)-1 do
  begin
    if (Lista[Index].Posicao.X = No.Posicao.X) and (Lista[Index].Posicao.Y = No.Posicao.Y) then
    begin
      Break;
    end;
  end;

  for i := Index + 1 to Length(Lista) - 1 do
  begin
    Lista[i -1 ] := Lista[i];
  end;
  SetLength(Lista, Length(Lista) - 1);
end;

function TfrmMain.CelulaMenorCusto(var Lista: TLista): TCelula;
var
  RIni, RFim: Integer;
  ListaIguais: TLista;
  MenorCusto: Double;
  idxMenorCusto, x, Index: Integer;
begin
  MenorCusto := MaxLongint;

  for x := 0 to Length(Lista) -1 do
  begin
    if Lista[x].Custo <= MenorCusto then
    begin
      MenorCusto := Lista[x].Custo;
      idxMenorCusto := x;
    end;
  end;

  for x := 0 to Length(Lista) -1 do
  begin
    if Lista[x].Custo = MenorCusto then
    begin
       Adicionar(ListaIguais, Lista[x]);
    end;
  end;

  //Se tiver vrios com o mesmo custo, ento escolhe aleatoriamente um deles
  if Length(ListaIguais) > 1 then
  begin
    Randomize;
    RFim := Length(ListaIguais);
    Index := Random(RFim);
  end
  else
    Index := 0;

  Result := ListaIguais[index];
end;

function TfrmMain.CelulaIgual(Celula1, Celula2: TCelula): Boolean;
begin
  Result := (Celula1.Posicao.X = Celula2.Posicao.X) and (Celula1.Posicao.Y = Celula2.Posicao.Y);
end;

function TfrmMain.EstaNaLista(var aLista: TLista; No: TCelula): Boolean;
var
  i: Integer;
  Achou: Boolean;
begin
  Achou := False;
  for i := 0 to Length(aLista) -1 do
  begin
    if CelulaIgual(aLista[i], No) then
    begin
      Achou := True;
      Break;
    end;
  end;

  Result := Achou;
end;

procedure TfrmMain.AtualizaCusto(var Atual: TCelula; const Inicial: TCelula; const Final: TCelula);
var
  HorI, VertI,
  HorF, VertF : Integer;
begin
  HorI := Inicial.Posicao.Y - Atual.Posicao.Y;
  if HorI < 0 then
    HorI := HorI *-1;

  VertI := Inicial.Posicao.X - Atual.Posicao.X;
  if VertI < 0 then
    VertI := VertI *-1;

  HorF := Final.Posicao.Y - Atual.Posicao.Y;
  if HorF < 0 then
    HorF := HorF *-1;

  VertF := Final.Posicao.X - Atual.Posicao.X;
  if VertF < 0 then
    VertF := VertF *-1;

  Atual.H := (HorF + VertF) * 100;  // Heurstica Prpria

  if not Atual.Diagonal then
    Atual.G := (HorI + VertI)
  else
    Atual.G := (HorI + VertI)*1.4142135623730950488016887242097; //Sqrt(2)

  Atual.Custo := (Atual.G) + Atual.H + Atual.CustoTerreno;
end;

procedure TfrmMain.ProcessarCelula(var Processar: TCelula; var Inicial: TCelula; var Final: TCelula; var Atual: TCelula; IndexPai: Integer);
var
  TaNaAberta,
  TaNaFechada: Boolean;
begin
  TaNaAberta := False;
  TaNaFechada := False;
  if Processar.Posicao.X > -1 then // -1  obstculo
  begin
    TaNaFechada := EstaNaLista(ListaFechada, Processar);

    if (not TaNaFechada) then
    begin
      TaNaAberta := EstaNaLista(ListaAberta, Processar);
      AtualizaCusto(Processar, Inicial, Final);
      if not TaNaAberta then
      begin
        Processar.Pai := IndexPai;
        Adicionar(ListaAberta, Processar);
      end
      else
      begin
        if Processar.G < Atual.G then
        begin
          Processar.Pai := IndexPai;
        end;
      end;
    end;
  end;
end;

procedure TfrmMain.AStar;
var
  ptrPai: Pointer;
  pPai: TCelula;
  i, j, x, y: Integer;
  MenorCusto,
  Start, Goal,
  vDireita, vEsquerda, vAcima, vAbaixo: TCelula;
  vSD, vSE, vID, vIE: TCelula;
begin
  SetLength(ListaFechada,0);
  SetLength(ListaAberta,0);
  SetLength(Caminho,0);

  for i := 0 to Mapa.ColCount -1 do
  begin
    for j := 0 to Mapa.RowCount -1 do
    begin
      if (AnsiUpperCase(Mapa.Cells[j, i]) = Cor_Visitado) or
         (AnsiUpperCase(Mapa.Cells[j, i]) = 'CLYELLOW') then
        Mapa.Cells[j, i] := '';
    end;
  end;
  Mapa.Repaint;

  AchouCaminho := False;
  lblStatus.Caption := 'Procurando...';
  lblStatus.Update;
  Start := GetStart;
  Goal := GetGoal;

  Start.CustoTerreno := 0;
  Start.Custo := 0;
  Start.H := 0;
  Start.G := -1;
  Start.Pai := -1;

  Goal.CustoTerreno := 0;
  Goal.Custo := 0;
  Goal.H := 0;
  Goal.G := 0;

  AtualizaCusto(Start, Start, Goal);

  CelulaAtual := Start;

  Adicionar(ListaAberta, Start);
  while (Length(ListaAberta) > 0) and (not AchouCaminho) do
  begin
    i := Length(Caminho);

    CelulaAtual := CelulaMenorCusto(ListaAberta);
    LimparDiagonais;

    if not EstaNaLista(Caminho, CelulaAtual) then
    begin
      SetLength(Caminho, i + 1);
      Caminho[i] := CelulaAtual;
    end;

    if i > 0 then
    begin
      CelulaAtual.Pai := i-1;
    end
    else
      CelulaAtual.Pai := -1;

    Adicionar(ListaFechada, CelulaAtual);
    Remover(ListaAberta, CelulaAtual);

    if CelulaIgual(CelulaAtual, Goal) then
    begin
      lblStatus.Caption := 'Cheguei!!!';
      AchouCaminho := True;
    end
    else
    begin
      if Mapa.Cells[CelulaAtual.Posicao.Y, CelulaAtual.Posicao.X] = '' then
        Mapa.Cells[CelulaAtual.Posicao.Y, CelulaAtual.Posicao.X] := Cor_Visitado;

      vDireita.Diagonal := False;
      vEsquerda.Diagonal := False;
      vAcima.Diagonal := False;
      vAbaixo.Diagonal := False;

      vSD.Diagonal := True;
      vSE.Diagonal := True;
      vID.Diagonal := True;
      vIE.Diagonal := True;

      VerificarVizinhos(vDireita, vEsquerda, vAcima, vAbaixo);
      VerificarDiagonais(vSD, vSE, vID, vIE);

      ProcessarCelula(vDireita, Start, Goal, CelulaAtual, i);
      ProcessarCelula(vEsquerda, Start, Goal, CelulaAtual, i);
      ProcessarCelula(vAcima, Start, Goal, CelulaAtual, i);
      ProcessarCelula(vAbaixo, Start, Goal, CelulaAtual, i);

      ProcessarCelula(vSD, Start, Goal, CelulaAtual, i);
      ProcessarCelula(vSE, Start, Goal, CelulaAtual, i);
      ProcessarCelula(vID, Start, Goal, CelulaAtual, i);
      ProcessarCelula(vIE, Start, Goal, CelulaAtual, i);
    end;
  end;
  if not AchouCaminho then
    lblStatus.Caption := 'Nenhum acesso foi encontrado!';
end;

procedure TfrmMain.BitBtn5Click(Sender: TObject);
begin
  AStar;
  if AchouCaminho then
    Animar;
end;

procedure TfrmMain.Animar;
var
  Anterior: TCelula;
  i, X, Y: Integer;
begin
  SetLength(MelhorCaminho,0);
  Anterior.Posicao.X := -1;

  CelulaAtual := Caminho[CelulaAtual.Pai];
  repeat
    Adicionar(MelhorCaminho, CelulaAtual);
    CelulaAtual := Caminho[CelulaAtual.Pai];
  until (CelulaAtual.Pai = -1);

  for i := Length(MelhorCaminho) -1 downto 0 do
  begin
    if Anterior.Posicao.X > -1 then
      Mapa.Cells[Anterior.Posicao.Y, Anterior.Posicao.X] := 'CLYELLOW';

    CelulaAtual := MelhorCaminho[i];
    X := CelulaAtual.Posicao.X;
    Y := CelulaAtual.Posicao.Y;

    Anterior.Posicao.X := X;
    Anterior.Posicao.Y := Y;

    if i > 0 then
    begin
      if Mapa.Cells[Y, X] = Cor_Visitado then
        Mapa.Cells[Y, X] := 'CLBLUE';
    end
    else
    begin
      if Mapa.Cells[Y, X] = Cor_Visitado then
        Mapa.Cells[Y, X] := 'CLYELLOW';
    end;

    if speVelocidade.Value < 200 then
      Sleep(201 - speVelocidade.Value);
    Application.ProcessMessages;
  end;
end;

procedure TfrmMain.LimparTudo;
var
  x, y: Integer;
begin
  for y := 0 to Mapa.ColCount -1 do
  begin
    for x := 0 to Mapa.RowCount -1 do
    begin
      Mapa.Cells[y,x] := '';
    end;
  end;
  Mapa.Repaint;
  SetLength(ListaFechada,0);
  SetLength(ListaAberta,0);
  SetLength(Caminho,0);
end;

procedure TfrmMain.Button1Click(Sender: TObject);
begin
  LimparTudo;
end;

function TfrmMain.DireitaInferiorLiberado: Boolean;
begin
  Result := (CelulaAtual.Posicao.Y < Mapa.ColCount -1) and
            (CelulaAtual.Posicao.X < Mapa.RowCount -1) and
            (
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y+1, CelulaAtual.Posicao.X+1]) = '') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y+1, CelulaAtual.Posicao.X+1]) = 'CLLIME') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y+1, CelulaAtual.Posicao.X+1]) = 'CLRED') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y+1, CelulaAtual.Posicao.X+1]) = 'CLYELLOW')
            );
end;

function TfrmMain.DireitaSuperiorLiberado: Boolean;
begin
  Result := (CelulaAtual.Posicao.Y < Mapa.ColCount -1) and
            (CelulaAtual.Posicao.X > 0) and
            (
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y+1, CelulaAtual.Posicao.X-1]) = '') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y+1, CelulaAtual.Posicao.X-1]) = 'CLLIME') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y+1, CelulaAtual.Posicao.X-1]) = 'CLRED') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y+1, CelulaAtual.Posicao.X-1]) = 'CLYELLOW')
            );
end;

function TfrmMain.EsquerdaInferiorLiberado: Boolean;
begin
  Result := (CelulaAtual.Posicao.Y > 0) and
            (CelulaAtual.Posicao.X < Mapa.RowCount -1) and
            (
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y-1, CelulaAtual.Posicao.X+1]) = '') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y-1, CelulaAtual.Posicao.X+1]) = 'CLLIME') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y-1, CelulaAtual.Posicao.X+1]) = 'CLRED') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y-1, CelulaAtual.Posicao.X+1]) = 'CLYELLOW')
            );
end;

function TfrmMain.EsquerdaSuperiorLiberado: Boolean;
begin
  Result := (CelulaAtual.Posicao.Y > 0) and
            (CelulaAtual.Posicao.X > 0) and
            (
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y-1, CelulaAtual.Posicao.X-1]) = '') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y-1, CelulaAtual.Posicao.X-1]) = 'CLLIME') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y-1, CelulaAtual.Posicao.X-1]) = 'CLRED') or
              (AnsiUpperCase(Mapa.Cells[CelulaAtual.Posicao.Y-1, CelulaAtual.Posicao.X-1]) = 'CLYELLOW')
            );
end;

procedure TfrmMain.BitBtn7Click(Sender: TObject);
begin
  CorSelecionada := pnlBarro.Color; 
end;

procedure TfrmMain.BitBtn6Click(Sender: TObject);
begin
  CorSelecionada := pnlAgua.Color;
end;

procedure TfrmMain.BitBtn8Click(Sender: TObject);
begin
  CorSelecionada := pnlBarreira.Color;
end;

procedure TfrmMain.BitBtn10Click(Sender: TObject);
begin
  CorSelecionada := pnlInicio.Color; 
end;

procedure TfrmMain.BitBtn9Click(Sender: TObject);
begin
  CorSelecionada := pnlDestino.Color;
end;

procedure TfrmMain.BitBtn11Click(Sender: TObject);
begin
  CorSelecionada := pnlLimpar.Color;
end;

procedure TfrmMain.BitBtn12Click(Sender: TObject);
var
  Modelo: TStrings;
  x, y: Integer;
begin
  Modelo := TStringList.Create;
  for x := 0 to Mapa.RowCount -1 do
  begin
    for y := 0 to Mapa.ColCount -1 do
    begin
      if (Mapa.Cells[y, x] = ColorToString(pnlBarro.Color)) or
         (Mapa.Cells[y, x] = ColorToString(pnlAgua.Color)) or
         (Mapa.Cells[y, x] = ColorToString(pnlBarreira.Color)) or
         (Mapa.Cells[y, x] = ColorToString(pnlInicio.Color)) or
         (Mapa.Cells[y, x] = ColorToString(pnlDestino.Color)) then
      begin
        Modelo.Add(IntToStr(y)+'|'+IntToStr(x)+'|'+Mapa.Cells[y, x]);
      end;
    end;
  end;
  if Modelo.Count > 0 then
    if SaveDialog1.Execute then
    begin
      Modelo.SaveToFile(SaveDialog1.FileName);
    end;
end;

procedure TfrmMain.BitBtn13Click(Sender: TObject);
var
  CelulaTmp: TStrings;
  Modelo: TStrings;
  i, x, y : Integer;
begin
  if OpenDialog1.Execute then
  begin
    LimparTudo;
    Modelo := TStringList.Create;
    CelulaTmp := TStringList.Create;
    Modelo.LoadFromFile(OpenDialog1.FileName);
    for i := 0 to Modelo.Count -1 do
    begin
      CelulaTmp := Explode(Modelo[i], '|');
      y := StrToInt(CelulaTmp[0]);
      x := StrToInt(CelulaTmp[1]);

      if Mapa.ColCount < y then
        Mapa.ColCount := y;
        
      if Mapa.RowCount < x then
        Mapa.RowCount := x;

      Mapa.Cells[y, x] := CelulaTmp[2];
    end;
  end;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  CorSelecionada := pnlLimpar.Color;
  speLinhas.Value := Mapa.RowCount;
  speColunas.Value := Mapa.ColCount;
  speTamanho.Value := Mapa.DefaultColWidth;
  Mapa.DefaultRowHeight := Mapa.DefaultColWidth;
end;

procedure TfrmMain.speLinhasChange(Sender: TObject);
begin
  Mapa.RowCount := speLinhas.Value;
end;

procedure TfrmMain.speColunasChange(Sender: TObject);
begin
  Mapa.ColCount := speColunas.Value;
end;

procedure TfrmMain.speTamanhoChange(Sender: TObject);
begin
  Mapa.DefaultColWidth := speTamanho.Value;
  Mapa.DefaultRowHeight := speTamanho.Value;
end;

procedure TfrmMain.LimparDiagonais;
var
  i: Integer;
begin
  try
    for i := 0 to Length(ListaAberta) -1 do
    begin
      if ListaAberta[i].Diagonal then
        Remover(ListaAberta, ListaAberta[i]);
    end;
  except
    on E: Exception do
    begin
      raise Exception.Create(IntToStr(i)+#13#10);
    end;
  end;
end;

end.
